/*
Package cmd defines user commands accessible from the cli tool.

Copyright © 2022 Joel D. Elkins <joel@elkins.co>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package cmd

import (
	"context"
	"fmt"
	"io"
	"os"

	"gitea.elkins.co/Networking/ccl/internal/pkg/config"
	"github.com/containers/podman/v4/pkg/bindings"
	log "github.com/sirupsen/logrus"
	"github.com/spf13/cobra"
	"golang.org/x/exp/slices"
)

// rootCmd represents the base command when called without any subcommands
var rootCmd = &cobra.Command{
	Use:     "ccl",
	Short:   "Manage a set of pre-configured of podman containers",
	Version: version,
	Long: `ccl is a utility to manage a set of podman containers. Use with various subcommands
to define, start, stop, or update the container images. Configuration is read
from a toml configuration file, and the utility uses this information to
execute the necessary podman commands.`,
	ValidArgsFunction: cobra.NoFileCompletions,
	PersistentPreRun: func(cmd *cobra.Command, args []string) {
		requireConn := []string{"create", "pull", "recreate", "restart", "rm", "start", "stop", "update"}
		if slices.Contains(requireConn, cmd.Name()) {
			// connect to podman
			connectMust()
		} else {
			if err := connect(); err != nil {
				log.WithField("error", err).Warnln("Connection failed")
			}
		}
		err := config.Init(conn)
		if err != nil {
			fmt.Fprintln(os.Stderr, "Warning: Could not initialize configuration:", err)
		} else if !incDisabled {
			for _, c := range config.Categories() {
				if len(c) > 0 && c[0] == '.' {
					contMask = append(contMask, c)
				}
			}
		}
	},
}

var (
	conn        context.Context
	output      io.Writer
	verbose     bool
	fake        bool
	socket      string
	incDisabled bool
	contMask    []string
)

// Execute adds all child commands to the root command and sets flags appropriately.
// This is called by main.main(). It only needs to happen once to the rootCmd.
func Execute() {
	err := rootCmd.Execute()
	if err != nil {
		os.Exit(1)
	}
}

func connect() error {
	var err error
	key := struct{ string }{"output"}
	conn, err = bindings.NewConnection(context.WithValue(context.Background(), key, output), socket)
	return err
}

func connectMust() {
	if err := connect(); err != nil {
		log.WithField("error", err).Errorf("Could not connect")
		os.Exit(1)
	}
}

func init() {
	cobra.OnInitialize(func() {
		if verbose {
			output = os.Stdout
		} else {
			output = io.Discard
		}
	})
	rootCmd.PersistentFlags().StringVarP(&config.ConfigFile, "config", "c", config.ConfigFileDefault, "pathname of config file")
	rootCmd.PersistentFlags().BoolVarP(&verbose, "verbose", "v", false, "show additional info from command execution")
	rootCmd.PersistentFlags().BoolVarP(&fake, "no-action", "n", false, "do not actually execute commands")
	rootCmd.PersistentFlags().BoolVarP(&incDisabled, "include-disabled", "a", false, "include category=. containers in actions")
	rootCmd.PersistentFlags().StringVarP(&socket, "socket", "k", "unix:///run/podman/podman.sock", "socket address to which to connect")
}
