/*
Package command is used to manage the container management instructions required to carry out
tasks desired by the user, such as "create"

Copyright © 2022 Joel D. Elkins <joel@elkins.co>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package command

import (
	"fmt"
	"io"
	"os/exec"
)

// Type is one of the Ct... constants, indicating the underlying method
// for executing a command.
type cType int

// Types of commands:
//   - CtNop: No operation; an empty placeholder
//   - CtShell: A shell command -- string to be fed to /bin/sh -c
//   - CtFunc: A go function of type `func() error`
//   - CtIndirect: A reference to another command. When executed, the underlying command is executed.
//   - CtSet: An ordered list of commands
//   - CtDebug: A debug message is printed upon execution, but does no other action
//   - CtConditional: A function and two commands are provided. When evaluated,
//     the function, having signature `func() bool`, is executed. If the result is
//     true, then the first command is executed, and if false the second one.
const (
	ctNop cType = iota
	ctShell
	ctFunc
	ctIndirect
	ctSet
	ctDebug
	ctConditional
)

func (ct cType) String() string {
	switch ct {
	case ctNop:
		return "NOP"
	case ctShell:
		return "SHELL"
	case ctFunc:
		return "FUNC"
	case ctIndirect:
		return "INDIRECT"
	case ctSet:
		return "SET"
	case ctDebug:
		return "DEBUG"
	case ctConditional:
		return "CONDITIONAL"
	default:
		return "UNKOWN"
	}
}

// Command combines a Type with an `interface{}`, representing a task (or a set
// of tasks, in the case of CtSet) to be executed. Create a Command with one of
// the `NewXxxCommand` functions.
type Command struct {
	Type    cType
	Command interface{}
}

// Commands reflects an ordered grouping of `Command`s
type Commands []Command

// Set bundes a `Commands` object with an arbitrary ID for verbose output
type Set struct {
	ID string
	Commands
}

type errFunc struct {
	Name string
	Func func() error
}

type conditional struct {
	Name      string
	Condition func() bool
	ThenCmd   Command
	ElseCmd   Command
}

// NewShell creates a new shell-type command from a string. Upon execution, the
// string will be fed to `/bin/sh -c`
func NewShell(cmd string) Command {
	return Command{ctShell, cmd}
}

// NewFunc creates a command consisting of go function of type `func() error`
func NewFunc(name string, f func() error) Command {
	return Command{ctFunc, errFunc{name, f}}
}

// NewIndirect creates a reference to another command. When executed, the underlying
// command is executed.
func NewIndirect(c Command) Command {
	return Command{ctIndirect, c}
}

// NewSet creates a single command containing an ordered list of commands. These
// commands will be executed in order with this set-type command is executed.
func NewSet(cs Set) Command {
	return Command{ctSet, cs}
}

// NewDebug returns a command that prints a debug message upon execution, but
// does no other action
func NewDebug(msg string) Command {
	return Command{ctDebug, msg}
}

// NewNop returns a placeholder command. When executed, nothing is done other
// than to note that it was encountered in the output.
func NewNop() Command {
	return Command{ctNop, nil}
}

// NewConditional takes a function and two commands as arguments. When evaluated,
// the function, having signature `func() bool`, is executed. If the result is
// true, then the first command is executed, and if false the second one.
func NewConditional(name string, ifPart func() bool, thenPart Command, elsePart Command) Command {
	return Command{ctConditional, conditional{
		Name:      name,
		Condition: ifPart,
		ThenCmd:   thenPart,
		ElseCmd:   elsePart,
	}}
}

// Execute the Set with the privileges of the user running the process.
// Warning: in the case of a shell command, this could do anything.
func (cmds Set) Execute(output io.Writer, fake bool) error {
	for _, c := range cmds.Commands {
		if err := c.Execute(output, fake, cmds.ID); err != nil {
			return err
		}
	}
	return nil
}

// Execute the Command with the privileges of the user running the process.
// Warning: in the case of a shell command, this could do anything.
func (c Command) Execute(output io.Writer, fake bool, commandSetID string) error {
	switch c.Type {
	case ctNop:
		fmt.Fprintf(output, "%s: %s\n", commandSetID, c.Type)
	case ctShell:
		cmd := c.Command.(string)
		fmt.Fprintf(output, "%s: %s sh -c \"%s\"\n", commandSetID, c.Type, cmd)
		if !fake {
			out, err := exec.Command("sh", "-c", cmd).CombinedOutput()
			fmt.Fprint(output, string(out))
			return err
		}
	case ctFunc:
		ef := c.Command.(errFunc)
		fmt.Fprintf(output, "%s: %s %s\n", commandSetID, c.Type, ef.Name)
		if !fake {
			if err := ef.Func(); err != nil {
				return err
			}
			return nil
		}
	case ctIndirect:
		ct := c.Command.(Command)
		return ct.Execute(output, fake, commandSetID+"/"+c.Type.String())
	case ctSet:
		cs := c.Command.(Set)
		for i := range cs.Commands {
			err := cs.Commands[i].Execute(output, fake, commandSetID+"/"+cs.ID)
			if err != nil {
				return err
			}
		}
	case ctConditional:
		cond := c.Command.(conditional)
		if fake {
			// in a fake setting, we don't know which branch will be followed,
			// so show what we would do in either branch
			fmt.Fprintf(output, "%s: %s %s\n", commandSetID, c.Type, cond.Name)
			cond.ThenCmd.Execute(output, fake, commandSetID+"/"+cond.Name+"/true")
			cond.ElseCmd.Execute(output, fake, commandSetID+"/"+cond.Name+"/false")
		} else {
			branch := cond.Condition()
			fmt.Fprintf(output, "%s: %s %s: %v\n", commandSetID, c.Type, cond.Name, branch)
			if branch {
				return cond.ThenCmd.Execute(output, fake, commandSetID+"/"+cond.Name+"/true")
			}
			return cond.ElseCmd.Execute(output, fake, commandSetID+"/"+cond.Name+"/false")
		}
	case ctDebug:
		msg := c.Command.(string)
		fmt.Fprintf(output, "%s: %s %s\n", commandSetID, c.Type, msg)
	}
	return nil
}
